/*************************************************************************************/
/* Workshop de
 * Supercomputación y Cálculo Científico, SC3-2014
 * SC3UIS con Academia NVIDIA CUDA
 
	Este codigo presenta un conjunto de kernels CUDA C del metodo Montecarlo
	para el calculo del numero pi. Los kernels aqui propuestos emplean ejecucion
	multiblock.
	
	Igualmente se presenta la generacion de numeros aleatorios no repetitivos sobre
	la GPU, para evitar problemas de periodicidad que no corresponden a la naturaleza
	probabilistica del metodo montecarlo.
	
	Finalmente se menciona que los kernels aqui presentados emplean declaraciones de 
	las librerias Thrust y CUDA C.
	
 * Creado por: William J Trigos, Ing. Electrónico, UIS(2013) 
 * Fecha: 20/04/2014 12:49 PM, (Bucaramanga, Colombia)
 */
/*************************************************************************************/

// Generacion de semillas aleatorias no repetitivas para generar numeros aleatorios en la GPU
__host__ __device__ unsigned int hash(unsigned int a){
a = (a+0x7ed55d16) + (a<<12);
a = (a^0xc761c23c) ^ (a>>19);
a = (a+0x165667b1) + (a<<5);
a = (a+0xd3a2646c) ^ (a<<9);
a = (a+0xfd7046c5) + (a<<3);
a = (a^0xb55a4f09) ^ (a>>16);
return a;
}

//////////////////////////////////////////////////////////////////////////////
//METODO 1: Empleando un kernel thrust mediante la funcion  
//////////////////////////////////////////////////////////////////////////////

//Kernel
//__global__ void piKernel(bool *count, int i, int ThreadsPerIteration)
__global__ void piKernel(bool *count, unsigned int shift){

unsigned int index = (blockIdx.x * blockDim.x) + threadIdx.x;

thrust::default_random_engine rng(hash(index + shift));
thrust::uniform_real_distribution<float> X(0,1);

float u = X(rng);
float v = X(rng);

if(u * u + v * v <= 1.0f)
	count[index] = true;
else
	count[index] = false;
}

// Invocacion desde el Host 
float pi_cuda(unsigned int tileSize, unsigned int threadsPerIteration){

unsigned int iterations = ceil(samples / (float)threadsPerIteration);					//Calcular las iteraciones a realizar, basados en el tamano de los threads lanzados
unsigned int blocksPerIteration = ceil(threadsPerIteration / (float)tileSize);			//Calcular Blocks

unsigned int nInside = 0;																//Initialize a counter
thrust::device_ptr<bool> CountPtr = thrust::device_malloc<bool>(threadsPerIteration);	//Crear un vector en el dispositivo para el conteo de puntos Initialize array to store values
bool *cPtr = thrust::raw_pointer_cast(CountPtr.get());									//Store pointer to first index

for(int i = 0; i < iterations; i++)														// ciclo iterativo
{
	piKernel<<<blocksPerIteration, tileSize>>>(cPtr, i * threadsPerIteration);			//llamar el Kernel
	nInside += thrust::count(CountPtr, CountPtr + threadsPerIteration, true);			//iniciar el conteo
}

thrust::device_free(CountPtr);															// liberar la memoria CUDA del puntero de conteo
return 4.0f * nInside / (float)(threadsPerIteration * iterations);						// regresar el valor de PI estimado
}

//////////////////////////////////////////////////////////////////////////////
//METODO 2: Haciendo uso de la operacion tranformacion de la libreria Thrust
//////////////////////////////////////////////////////////////////////////////

struct Find_PI{
__host__ __device__ bool operator()(unsigned int thread_id)
	{
		thrust::default_random_engine rng(hash(thread_id));
		thrust::uniform_real_distribution<float> X(0,1);
		float u = X(rng);
		float v = X(rng);
		if(u * u + v * v <= 1.0f)
			return true;
		else
			return false;
	}
};

// Invocacion desde el Host
float pi_cuda2(){
float pi_sum = thrust::transform_reduce(thrust::counting_iterator<int>(0),	//Primera posicion del vector
							thrust::counting_iterator<int>(samples),		//Ultima posicion del vector
							Find_PI(),										//Funcion a ejecutar
							0.0f,											//Valor Inicial de pi_sum
							thrust::plus<float>()							//Add function
							);
return 4.0f * pi_sum / samples;
}

//////////////////////////////////////////////////////////////////////////////
//METODO 3: Haciendo uso CUDA neto, con memoria global
//////////////////////////////////////////////////////////////////////////////

// Kernel that executes on the CUDA device
__global__ void cal_pi(float *sum, int nbin, float step, int nthreads, int nblocks) {
	int i;
	float x;
	// Identificador Secuencial de cada thread a lo largo de los blocks
	int idx = blockIdx.x*blockDim.x + threadIdx.x; 
	for (i=idx; i< nbin; i+=nthreads*nblocks) {
		x = (i+0.5)*step;
		sum[idx] += 4.0/(1.0+x*x);
	}
}

__device__ float my_rand(unsigned int *seed) {
	unsigned long a = 16807; // constants for random number generator
	unsigned long m = 2147483647;
	// 2^31 - 1
	unsigned long x = (unsigned long) *seed;
	x = (a * x)%m;
	*seed = (unsigned int) x;
	return ((float)x)/m;
}

// Kernel that executes on the CUDA device
__global__ void cal_pi5(float *sum, int nthreads, int nblocks) {

unsigned int idx = threadIdx.x + blockDim.x * blockIdx.x;
int points_in_circle = 0;
float x, y;
unsigned int seed = idx + 1;
// starting number in random sequence
	for(int i = 0; i < nblocks*nthreads; i++) {
		x = my_rand(&seed);
		y = my_rand(&seed);
		points_in_circle += (x*x + y*y <= 1.0f); // count if x & y is in the circle.
	}
	sum[idx] = 4.0f * points_in_circle / (float) (nblocks*nthreads);
}

float pi_cuda3(unsigned int tileSize, unsigned int threadsPerIteration){
	#define NBIN samples  
	unsigned int NUM_BLOCK = ceil(threadsPerIteration / (float)tileSize);	//Calcular Blocks
	unsigned int NUM_THREAD = tileSize;										// Dimension de los Bloques
	dim3 dimGrid(NUM_BLOCK,1,1);  											// Dimension de la Grid ()
	dim3 dimBlock(NUM_THREAD,1,1);											// Dimension de los Bloques
	float *sumHost, *sumDev;										// Punteros de los vectores en host & GPU (device) 

	float step = 1.0/NBIN;  												// Tamaño de paso
	float pi = 0.0f;
	
	size_t size = NUM_BLOCK*NUM_THREAD*sizeof(float);  						// Tamano del vector
	sumHost = (float *)malloc(size);  										// Crear vector host
	cudaMalloc((void **) &sumDev, size);  									// Crear vector GPU (device)

	// Inicializar el vector en GPU a valores "0"
	cudaMemset(sumDev, 0, size);

	// Ejecutar kernel en GPU
	cal_pi <<<dimGrid, dimBlock>>> (sumDev, NBIN, step, NUM_THREAD, NUM_BLOCK);

	// Tranferir resultados desde GPU a CPU
	cudaMemcpy(sumHost, sumDev, size, cudaMemcpyDeviceToHost);
	
	// Porsion serial en CPU
	for(int tid=0; tid<NUM_THREAD*NUM_BLOCK; tid++){
		pi += sumHost[tid];
	}
	pi *= step;

	// Imprimir resultado
	//printf("PI = %.12f\n",pi);

	// Liberar memoria
	free(sumHost); 
	cudaFree(sumDev);

	return pi;
	//return 0;
}

float pi_cuda5(unsigned int tileSize, unsigned int threadsPerIteration){

	unsigned int NUM_BLOCK = ceil(threadsPerIteration / (float)tileSize);	//Calcular Blocks
	unsigned int NUM_THREAD = tileSize;										// Dimension de los Bloques

	dim3 dimGrid(NUM_BLOCK,1,1);  											// Dimension de la Grid ()
	dim3 dimBlock(NUM_THREAD,1,1);											// Dimension de los Bloques

	float *sumHost, *sumDev;										// Punteros de los vectores en host & GPU (device) 
	float pi = 0.0f;

	size_t size = NUM_BLOCK*NUM_THREAD*sizeof(float);  						// Tamano del vector
	sumHost = (float *)malloc(size);  										// Crear vector host
	cudaMalloc((void **) &sumDev, size);  									// Crear vector GPU (device)

	// Inicializar el vector en GPU a valores "0"
	cudaMemset(sumDev, 0, size);

	// Ejecutar kernel en GPU
	cal_pi5 <<<dimGrid, dimBlock>>> (sumDev, NUM_THREAD, NUM_BLOCK);

	// Tranferir resultados desde GPU a CPU
	cudaMemcpy(sumHost, sumDev, size, cudaMemcpyDeviceToHost);
	
	// Porsion serial en CPU
	for(int tid=0; tid<NUM_THREAD*NUM_BLOCK; tid++){
		pi += sumHost[tid];
	}
	pi /= (NUM_THREAD*NUM_BLOCK);

	// Imprimir resultado
	//printf("PI = %.12f\n",pi);

	// Liberar memoria
	free(sumHost); 
	cudaFree(sumDev);

	return pi;
	//return 0;
}
