#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <cuda.h>
#include <curand_kernel.h>

// Desafio #1: encontrar las iteraciones necesarias para que (error < 6e-5)
//extern const long ITER_POR_THREAD = 5000;

#define BLOCKS 256
#define THREADS 256

// Desafio 2: Modificando BLOCKS y THREADS para mirar el efecto de emplear mas unidades de
// 				computo paralelo (mas nucleos de la GPU para calcular PI)

// para trabajar el desafio 2, descomente las lineas bajo este comentario y comente la linea 9
#define MAX_PUNTOS 327680000//655360000
extern const long ITER_POR_THREAD = (MAX_PUNTOS)/(BLOCKS*THREADS);

#define PI 3.1415926535 		// Valor de pi segun google

// Creando el kernel de Montecarlo para el lenguaje CUDA
__global__ void gpu_monte_carlo(float *estimate, curandState *states) {
	unsigned int tid = threadIdx.x + blockDim.x * blockIdx.x;
	long points_in_circle = 0;
	float x, y;
	curand_init(1234, tid, 0, &states[tid]); // Inicializacion de la libreria CURAND

	for(int i = 0; i < ITER_POR_THREAD; i++) {
		x = curand_uniform (&states[tid]);
		y = curand_uniform (&states[tid]);
		points_in_circle += (x*x + y*y <= 1.0f); // verificar si los puntos estan dentro de r=1
	}
	estimate[tid] = 4.0f*points_in_circle / (float) ITER_POR_THREAD; //aproximando pi
}

// funcion de Montecarlo que corre sobre CPU, en lenguaje C
float host_monte_carlo(long long iter) {
	float x, y;
	long points_in_circle;
	for(long i = 0; i < iter; i++) {
		x = rand() / (float) RAND_MAX;
		y = rand() / (float) RAND_MAX;
		points_in_circle += (x*x + y*y <= 1.0f);
	}
	float result = 4.0f*points_in_circle/iter;
	return result;
}

// Programa Principal
int main (int argc, char *argv[]) {	
	clock_t start, stop;
	int runs = 3;
	float host[BLOCKS * THREADS];
	float *dev;
	curandState *devStates;

		printf("*********************************************************************\n");
		printf("	Calculo de PI usando Monte Carlo con %ld puntos    \n", (ITER_POR_THREAD*THREADS*BLOCKS));
		printf("  #itera x thread = %ld, # blocks = %d, # threads/block = %d.\n\n", ITER_POR_THREAD, BLOCKS, THREADS);
		printf("*********************************************************************\n");

	for(int j=1; j<=runs; j++){
		start = clock();
		
		// reservando memoria en GPU
		cudaMalloc((void **) &dev, BLOCKS * THREADS * sizeof(float));
		cudaMalloc( (void **)&devStates, THREADS * BLOCKS * sizeof(curandState) );
		
		// lanzamiento de trabajo sobre GPU (invocacion de kernel CUDA)
		gpu_monte_carlo<<<BLOCKS, THREADS>>>(dev, devStates);

		// copiado hacia host de los resultados "Parciales" obtenidos en GPU
		cudaMemcpy(host, dev, BLOCKS * THREADS * sizeof(float), cudaMemcpyDeviceToHost);
		
		/* Efectuando una operacion de reduccion para la obtencion del numero pi */
		float pi_gpu;
		for(int i = 0; i < BLOCKS * THREADS; i++) {
			pi_gpu += host[i];
		}
		pi_gpu /= (BLOCKS * THREADS);
		stop = clock();


		/* IMPRIMIENDO LOS RESULTADOS DE LA EJECUCION */
		printf("\n Experimento # %d", j);
		printf("\n*********************************************************************\n");
		printf("\n 	Tiempo para calcular en GPU [%0.3e s].\n", (stop-start)/(float)CLOCKS_PER_SEC);
		
		start = clock();
		float pi_cpu = host_monte_carlo((long)BLOCKS * (long)THREADS * (long)ITER_POR_THREAD);
		stop = clock();
		// Imprimiendo Resultados de la ejecucion
		printf(" 	Tiempo para calcular en CPU [%0.3e s].\n", (stop-start)/(float)CLOCKS_PER_SEC);

		printf(" 	CUDA PI = %0.9f [error %0.3e]\n", pi_gpu, pi_gpu - PI);
		printf(" 	CPU  PI = %0.9f [error %0.3e]\n", pi_cpu, pi_cpu - PI);
	}
	return 0;
}
